@@ -1,18 +1,9 @@
-# /bin/bash
+#! /bin/bash
 # Functions used by test scripts.
 
-# programs used:
-# Xvfb (package xvfb)
-# xauth (package xauth)
-# mcookie (package util-linux)
-# xwininfo, xmessage (package x11-utils)
-# metacity (package metacity) if $USE_WM is set
-
 # Side effect: sets variables logdir and xlbiff_to_test
 parse_command_line() {
-    cd $(dirname "$0")/.. || exit 1
     logdir=
-    xlbiff_to_test=./xlbiff
     while [[ $# -gt 0 ]]; do
         case $1 in
             --logdir)
@@ -23,13 +14,8 @@ parse_command_line() {
                 logdir=$2
                 shift
                 ;;
-            --binary)
-                if [[ $# -lt 2 ]]; then
-                    echo "$0: --binary requires an argument" >&2
-                    exit 1
-                fi
-                xlbiff_to_test=$2
-                shift
+            --as-installed)
+                as_installed=1
                 ;;
             -*)
                 echo "$0: unknown option: $1" >&2
@@ -42,8 +28,55 @@ parse_command_line() {
         esac
         shift
     done
+    if [[ -n "$as_installed" ]]; then
+        # change out of the source package root, so we don't accidentally
+        # use resource and form files there when we think we are using
+        # installed versions.
+        cd "$(dirname "$0")" || exit 1
+        xlbiff_to_test=/usr/bin/xlbiff        
+        resource_file=/etc/X11/app-defaults/XLbiff
+    else
+        # as built
+        xlbiff_to_test=./xlbiff
+        resource_file=./XLbiff
+        PATH=.:/bin:/usr/bin
+    fi
 }
 
+# programs used:
+
+# if $NO_XVFB is not set
+xvfb_dependencies=(
+    Xvfb xvfb
+    xauth xauth
+    mcookie util-linux
+    xwininfo x11-utils
+    xmessage x11-utils
+)
+
+# if $USE_WM is set
+# Choice of WM is arbitrary; metacity is simple to run
+wm_dependency=(
+    metacity metacity
+)
+
+util_check_dependencies() {
+    # arguments are alternating binary and package
+    local binary package
+    while (($# > 0)); do
+        binary=$1
+        package=$2
+        shift; shift
+        if [[ -z "$(type -p "$binary")" ]]; then
+            echo "$0: program not available: $binary" >&2
+            echo "$0: package \"$package\" is probably not installed" >&2
+            echo "$0: (not required for build, only for this test)" >&2
+            exit 2
+        fi
+    done
+}
+
+
 # Creates a temp directory and sets an exit trap to remove it.
 # Side effects:
 # Sets variable test_tmpdir.
@@ -78,24 +111,24 @@ start_xlbiff_under_xvfb() {
     export TMPDIR="$test_tmpdir"
     PATH=$test_tmpdir:$PATH
 
+    util_logv "$(dpkg-query --show \
+                 -f 'Package ${Package} ${Version} ${Status}' xlbiff 2>&1)"
+
     if [[ -n "$NO_XVFB" ]]; then
+        util_logv "$("$xlbiff_to_test" -version 2>&1)"
+        util_logv "$xlbiff_to_test" "$@"
         timeout 60 "$xlbiff_to_test" "$@" &
         xvfb_pid=$!
         return
     fi
 
-    local xvfb_binary=/usr/bin/Xvfb
-    if [[ ! -x "$xvfb_binary" ]]; then
-        echo "$0: program not available: $xvfb_binary" >&2
-        echo "$0: package xvfb is probably not installed" >&2
-        echo "$0: (not required for build, only for this test)" >&2
-        exit 2
-    fi
+    util_check_dependencies "${xvfb_dependencies[@]}"
     local -i display_num=19
     # loop looking for an available display number
     while ((++display_num < 64)); do
         kill_xvfb
-        local xauth_cookie=$(mcookie)
+        local xauth_cookie
+        xauth_cookie=$(mcookie)
         export XAUTHORITY="$test_tmpdir/Xauthority"
         echo -n > "$XAUTHORITY"
         export DISPLAY=":$display_num"
@@ -106,8 +139,8 @@ start_xlbiff_under_xvfb() {
         trap : USR1
         # Pass SIGUSR1 as ignored, so Xvfb will signal its parent when ready.
         # See xserver(1).
-        (trap '' USR1; exec "$xvfb_binary" "$DISPLAY" -auth "$XAUTHORITY" \
-                 2> "$logdir/xvfb.$current_test_name.log") &
+        (trap '' USR1; exec Xvfb "$DISPLAY" -auth "$XAUTHORITY" \
+                 2>> "$logdir/xvfb.$current_test_name.log") &
         xvfb_pid=$!
 
         # Wait for Xvfb to to either signal us (if ready) or die (if
@@ -121,9 +154,14 @@ start_xlbiff_under_xvfb() {
 
         # Start the first client
         if [[ -n "$USE_WM" ]]; then
-            # Choice of WM is arbitrary; metacity is simple to run
-            TEST_WM="${TEST_WM:-metacity}"
-            [[ -n "$VERBOSE" ]] && echo "$0: starting window manager $TEST_WM"
+            if [[ -n "$TEST_WM" ]]; then
+                local probable_wm_pkg_name="$(basename "$TEST_WM")"
+                util_check_dependencies "$TEST_WM" "$probable_wm_pkg_name"
+            else
+                util_check_dependencies "${wm_dependency[@]}"
+                TEST_WM="${wm_dependency[0]}"
+            fi
+            util_logv "starting window manager $TEST_WM"
             "$TEST_WM" 2>> "$logdir/xvfb.$current_test_name.log" &
             util_wait_for_client_has_connected "Sawfish|Metacity|$TEST_WM" \
                 || continue
@@ -135,7 +173,9 @@ start_xlbiff_under_xvfb() {
         fi
 
         # everything started up successfully
-        "$xlbiff_to_test" "$@" &
+        util_logv "$("$xlbiff_to_test" -version 2>&1)"
+        util_logv "$xlbiff_to_test" "$@"
+        "$xlbiff_to_test" "$@" >> "$logdir/xvfb.$current_test_name.log" 2>&1 &
         xlbiff_pid=$!
         return
     done
@@ -149,7 +189,18 @@ util_wait_for_client_has_connected() {
 }
 
 util_client_has_connected() {
-    xwininfo -root -children 2>&1 | egrep -q -i "$util_window_name_pattern"
+    xwininfo -root -children 2>&1 | grep -E -q -i "$util_window_name_pattern"
+}
+
+# append to the log, and if verbose to stdout.
+# prepend the messages with program name and time.
+util_logv() {
+    local prefix
+    prefix="$0 $(date +'%H:%M:%S.%3N')"
+    echo "$prefix" "$@" >> "$logdir/xvfb.$current_test_name.log"
+    if [[ -n "$VERBOSE" ]]; then
+        echo "$prefix" "$@"
+    fi
 }
 
 # Returns success if test with this name should be run.
@@ -172,7 +223,9 @@ start_test() {
     fi
     current_test_name=$test_name
     ((num_tests_run++))
-    [[ -n "$VERBOSE" ]] && echo "$0: starting test $current_test_name"
+    mv "$logdir/xvfb.$current_test_name.log" \
+       "$logdir/xvfb.$current_test_name.log.prev" 2>/dev/null
+    util_logv "starting test $current_test_name"
     return 0
 }
 
@@ -204,9 +257,11 @@ loop_for() {
     local return_on_failure="$4"
 
     local child_alive=1
+    local -i loops_done=0
     while sleep 0.1; do
-        if ((--loop_count <= 0)); then
-            echo "$0: timed out waiting for $success_function" \
+        if ((++loops_done > loop_count)); then
+            echo "$0 $(date +'%H:%M:%S.%3N') timed out" \
+                 "after $loop_count tries waiting for $success_function" \
                  "in test $current_test_name $context_msg" >&2
             xauth -v -i -n list >&2
             xwininfo -root -tree >&2
@@ -218,16 +273,17 @@ loop_for() {
             child_alive=0
         fi
         if $success_function; then
-            [[ -n "$VERBOSE" ]] && echo "$0: $success_function true"
+            util_logv "try $loops_done/$loop_count: $success_function true"
             break
         fi
         if [[ "$child_alive" = 0 ]]; then
-            echo "$0: child process has died waiting for $success_function" \
+            echo "$0: child process has died on try $loops_done/$loop_count" \
+                 "waiting for $success_function" \
                  "in test $current_test_name $context_msg" >&2
             [[ -n "$return_on_failure" ]] && return 1
             exit 1
         fi
-        [[ -n "$VERBOSE" ]] && echo "$0: $success_function false"
+        util_logv "try $loops_done/$loop_count: $success_function false"
     done
 }
 
